import cv2
import mediapipe as mp
from sixdrepnet import SixDRepNet
from face_detection import RetinaFace
from scipy.spatial import distance
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter
import numpy as np
import time
import json
import logging
import math
from .utils import (
    plot_pose_cube,
    filter_video_data,
    plot_head_data_mov,
    plot_head_data_speed
)
from .utils import ensure_dir_created

logging.basicConfig(format='APOLLO: (%(asctime)s): %(message)s', datefmt='%d/%m/%Y %I:%M:%S %p',
                    level=logging.DEBUG)


def mouth_movement(filepath: str, focal_length_mm: int, output_dir='./output', plot=False,
                   normalize=False, lim: list[float, float] = None) -> str:
    """
    Generates the face data points for a given video, it currently accepts a small variety of part indices in the face

    :param filepath:
    :param focal_length_mm:
    :param output_dir:
    :param plot:
    :param normalize:
    :param lim:
    :return:
    """
    ensure_dir_created(output_dir)

    file_direction = output_dir + "/" + filepath.split("/")[-1].split(".")[0]

    distances_data = []

    video_source = cv2.VideoCapture(filepath)
    mp_face_dots = mp.solutions.face_mesh

    fps_general = int(video_source.get(cv2.CAP_PROP_FPS))
    frame_width = int(video_source.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(video_source.get(cv2.CAP_PROP_FRAME_HEIGHT))
    total_frames = int(video_source.get(cv2.CAP_PROP_FRAME_COUNT))

    start_frame, end_frame = 0, total_frames

    # Check if the limits are correctly set if defined
    if lim is not None:
        start_frame, end_frame = [int(x * fps_general) for x in lim]
        if start_frame < 0 or end_frame > total_frames:
            raise ValueError("The time limits do not match with the length of the video")

    # Codec and VideoWriter object to save the output video
    fourcc = cv2.VideoWriter_fourcc(*'XVID')

    out = cv2.VideoWriter(f'{file_direction}_face_points.mp4', fourcc, fps_general,
                          (frame_width, frame_height))

    # Mediapipe face landmark with 468 points
    face_mesh = mp_face_dots.FaceMesh(min_detection_confidence=0.5)

    max_distance = 0  # Initialize the maximum observed distance
    min_distance = float('inf')  # Initialize the minimum observed distance

    frame_counter = 0

    while video_source.isOpened():
        ret, frame = video_source.read()

        # If not more frames are found stop the recognition
        if not ret or frame is None:
            break

        # Skip frames outside limits
        if lim is not None and (frame_counter < start_frame or frame_counter > end_frame):
            frame_counter += 1
            continue

        rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        results_facial_dots = face_mesh.process(rgb_frame)

        fh, fw, _ = frame.shape

        if results_facial_dots.multi_face_landmarks:
            # Only select the points 0 and 17
            index_top, index_bottom, mp_face_part = 0, 17, mp_face_dots.FACEMESH_LIPS

            for source_idx, _ in mp_face_part:
                source = results_facial_dots.multi_face_landmarks[0].landmark[source_idx]
                x = int(source.x * fw)
                y = int(source.y * fh)

                source_top = results_facial_dots.multi_face_landmarks[0].landmark[index_top]
                x_top = int(source_top.x * fw)
                y_top = int(source_top.y * fh)

                source_bottom = results_facial_dots.multi_face_landmarks[0].landmark[index_bottom]
                x_bottom = int(source_bottom.x * fw)
                y_bottom = int(source_bottom.y * fh)

                # Applying chebyshev distance
                distance_px = int(distance.chebyshev([x_top, y_top], [x_bottom, y_bottom]))
                distance_mm = int(
                    (distance_px * focal_length_mm) / (focal_length_mm * 2 * math.tan(math.radians(0.5))))
                timestamp = video_source.get(cv2.CAP_PROP_POS_MSEC)
                distances_data.append([(distance_mm / 100), timestamp / 1000])

                cv2.circle(frame, (x, y), 2, (255, 255, 255), -1)

                max_distance = max(max_distance, distance_mm / 100)
                min_distance = min(min_distance, distance_mm / 100)

        out.write(frame)  # Write the frame to the output video file
        frame_counter += 1

    video_source.release()
    cv2.destroyAllWindows()

    if len(distances_data) <= 0:
        logging.info("Video not found, check directory")
        return "Video not found"

    logging.info("Video found, starting processing...")

    distances_data = filter_video_data(distances_data, normalize)
    x_values = distances_data['Timestamp'].tolist()
    y_values = distances_data['Distance'].tolist()

    data = {"x_values": x_values, "y_values": y_values, "fps": fps_general, "min_distance": min_distance,
            "max_distance": max_distance, "frame_width": frame_width, "frame_height": frame_height}

    if plot:
        # Middle line value to specify the mouth closing region
        threshold = ((max_distance + min_distance) / 2) * 0.9

        plt.figure(figsize=(10, 6))
        plt.plot(x_values, y_values, label="Part Opening Distance", color='b')
        plt.xlabel("Timestamp (s)")
        plt.ylabel("Distance (mm)")
        plt.title("Distance vs. Timestamp")
        plt.axhline(y=threshold, color='r', linestyle='--', label='Threshold')
        plt.grid(True)
        plt.legend()
        plt.savefig(f'{file_direction}.png')

    return json.dumps(data)


def head_position(filepath: str, output_dir="./output", plot=False, real_head_width=145,
                  lim: list[float, float] = None) -> str:
    """
    Estimates the head position over time

    :param filepath:
    :param output_dir:
    :param plot:
    :param real_head_width:
    :param lim:
    :return:
    """
    ensure_dir_created(output_dir)

    file_direction = output_dir + "/" + filepath.split("/")[-1].split(".")[0]

    model = SixDRepNet()
    detector = RetinaFace()

    video_source = cv2.VideoCapture(filepath)
    logging.info("Video found, starting processing...")

    fps_general = int(video_source.get(cv2.CAP_PROP_FPS))
    frame_width = int(video_source.get(cv2.CAP_PROP_FRAME_WIDTH))
    frame_height = int(video_source.get(cv2.CAP_PROP_FRAME_HEIGHT))
    total_frames = int(video_source.get(cv2.CAP_PROP_FRAME_COUNT))

    start_frame, end_frame = 0, total_frames

    # Check if the limits are correctly set if defined
    if lim is not None:
        start_frame, end_frame = [int(x * fps_general) for x in lim]
        if start_frame < 0 or end_frame > total_frames:
            raise ValueError("The time limits do not match with the length of the video")

    # Codec and VideoWriter object to save the output video
    fourcc = cv2.VideoWriter_fourcc(*'XVID')

    out = cv2.VideoWriter(f'{file_direction}_head_position.mp4', fourcc,
                          fps_general,
                          (frame_width, frame_height))

    prev_time = None
    prev_head_pos = None
    head_movement_speeds = []
    yaws = []
    pitchs = []
    rolls = []

    start_time = time.time()
    frame_counter = 0
    bbox_width = 0

    while video_source.isOpened():
        ret, frame = video_source.read()

        # If not more frames are found stop the recognition
        if not ret or frame is None:
            break

        # Skip frames outside limits
        if lim is not None and (frame_counter < start_frame or frame_counter > end_frame):
            frame_counter += 1
            continue

        faces = detector(frame)

        if len(faces) == 0:
            # No faces detected, add zeros
            yaws.append(0)
            pitchs.append(0)
            rolls.append(0)
            head_movement_speeds.append(0)
            continue

        for box, landmarks, score in faces:
            if score < .95:
                continue
            x_min = int(box[0])
            y_min = int(box[1])
            x_max = int(box[2])
            y_max = int(box[3])
            bbox_width = abs(x_max - x_min)
            bbox_height = abs(y_max - y_min)

            x_min = max(0, x_min - int(0.2 * bbox_height))
            y_min = max(0, y_min - int(0.2 * bbox_width))
            x_max = x_max + int(0.2 * bbox_height)
            y_max = y_max + int(0.2 * bbox_width)

            if x_min >= x_max or y_min >= y_max:
                logging.error("Invalid bounding box coordinates")
                continue

            img = frame[y_min:y_max, x_min:x_max]

            if frame.size == 0:
                logging.error("Empty image slice")
                continue

            pitch, yaw, roll = model.predict(img)

            if isinstance(pitch, (list, np.ndarray)):
                pitch = pitch[0]
            if isinstance(yaw, (list, np.ndarray)):
                yaw = yaw[0]
            if isinstance(roll, (list, np.ndarray)):
                roll = roll[0]

            yaws.append(yaw)
            pitchs.append(pitch)
            rolls.append(roll)

            curr_time = time.time()
            curr_head_pos = np.array([pitch, yaw, roll])
            if prev_head_pos is not None:
                head_movement = np.linalg.norm(curr_head_pos - prev_head_pos)
                time_elapsed = curr_time - prev_time
                head_movement_speed = head_movement / time_elapsed
                head_movement_speeds.append(head_movement_speed)

            prev_time = curr_time
            prev_head_pos = curr_head_pos

            frame = plot_pose_cube(frame, yaw, pitch, roll, x_min + int(.5 * (
                    x_max - x_min)), y_min + int(.5 * (y_max - y_min)), size=bbox_width)

        logging.info(f"Reading frame {frame_counter + 1}/{total_frames}")
        out.write(frame)
        frame_counter += 1

    logging.info(f"Time taken: {time.time() - start_time}")
    video_source.release()
    cv2.destroyAllWindows()

    # Formula to scale the real head with
    time_in_seconds = [frame_num / fps_general for frame_num in range(len(head_movement_speeds))]
    scale_factor = real_head_width / bbox_width

    head_speeds_mm_per_sec = [speed * scale_factor for speed in head_movement_speeds]

    # Calculate the average movement (this value can be compared to normal movement people)
    non_zero_speeds = [speed for speed in head_speeds_mm_per_sec if speed != 0]
    mean_speed = np.mean(non_zero_speeds)
    std_speed = np.std(non_zero_speeds)
    threshold = mean_speed + 2 * std_speed
    threshold_mm = threshold * scale_factor

    # Export x and y values for API (pitch, yaw, and roll values)
    min_length = min(len(time_in_seconds), len(pitchs), len(yaws), len(rolls))
    time_in_seconds = time_in_seconds[:min_length]
    pitch_vals = pitchs[:min_length]
    yaw_vals = yaws[:min_length]
    roll_vals = rolls[:min_length]

    # Export x and y values for API (head speed)
    x_head_speed = np.linspace(min(time_in_seconds), max(time_in_seconds), len(time_in_seconds))
    head_speeds_smooth = savgol_filter(head_speeds_mm_per_sec, 10, 3)  # Smooth with filter

    if plot:
        plot_head_data_mov(file_direction=file_direction,
                           time_in_seconds=time_in_seconds,
                           pitch_vals=pitch_vals,
                           yaw_vals=yaw_vals,
                           roll_vals=roll_vals)

        plot_head_data_speed(file_direction=file_direction,
                             time_in_seconds=x_head_speed,
                             head_speeds_mm=head_speeds_smooth,
                             threshold_mm=threshold_mm)

    data = {"x_head_speed": x_head_speed.tolist(),
            "y_head_speed": head_speeds_smooth.tolist(),
            "x_head_movement": time_in_seconds,
            "y_pitch_values": pitch_vals,
            "y_yaw_values": yaw_vals,
            "y_roll_values": roll_vals,
            "threshold_speed": float(threshold_mm),
            "average_speed": float(mean_speed)}

    return json.dumps(str(data))  # Does not know how it will work later




import numpy as np
import os
from scipy.io.wavfile import write
import matplotlib.pyplot as plt
import librosa


def save_to_file(filename: str, voice_data: list, samplerate: int) -> None:
    """
    Takes the filename (include directory if necessary) and the voice_data as a Queue
    including the samplerate to save the audio to wav

    :param filename:
    :param voice_data:
    :param samplerate:
    :return:
    """
    try:
        if not len(voice_data) == 0:
            data = np.concatenate(voice_data, axis=0)
            write(filename, samplerate, data)
    except Exception as e:
        raise Exception("Error saving file: {0}".format(e))


def lim_y_trim(sr: float, y: np.ndarray, lim: list[float, float] = None) -> np.ndarray:
    """
    Protected method to do the checking of the limiters in all sound analysis methods
    Do not use.

    :param sr:
    :param y:
    :param lim:
    :return:
    """
    if len(lim) != 2:
        raise ValueError("lim parameter: It only accepts two items, the start_time and end_time in seconds.")

    start_index = int(lim[0] * sr)
    end_index = int(lim[1] * sr)
    y = y[start_index:end_index]

    return y


def ensure_dir_created(output_dir: str) -> None:
    """
    Util function to ensure directory has been created and if
    not then create such directory

    :param output_dir:
    :return:
    """
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)


def plot_chroma(file_direction, s, chroma):
    """
    Plots chroma values

    :param file_direction:
    :param s:
    :param chroma:
    :return:
    """
    fig, ax = plt.subplots(nrows=2, sharex=True)
    img = librosa.display.specshow(librosa.amplitude_to_db(s, ref=np.max),
                                   y_axis='log', x_axis='time', ax=ax[0])
    fig.colorbar(img, ax=[ax[0]])
    ax[0].label_outer()
    img = librosa.display.specshow(chroma, y_axis='chroma', x_axis='time', ax=ax[1])
    fig.colorbar(img, ax=[ax[1]])
    plt.legend()
    plt.savefig(f'{file_direction}_chroma_features.png')


def plot_spectral_contrast(file_direction, contrast):
    """
    Plots spectral contrast

    :param file_direction:
    :param contrast:
    :return:
    """
    plt.figure(figsize=(10, 4))
    librosa.display.specshow(contrast, x_axis='time')
    plt.colorbar(format='%+2.0f dB')
    plt.title('Spectral contrast')
    plt.tight_layout()
    plt.legend()
    plt.savefig(f'{file_direction}_spectral_features.png')


def plot_onsets(file_direction, onset_env, onsets, times):
    """
    Plot onset detection

    :param file_direction:
    :param onset_env:
    :param onsets:
    :param times:
    :return:
    """
    plt.figure(figsize=(10, 5))
    plt.plot(times, onset_env / onset_env.max(), label='Normalized Onset Strength')
    plt.vlines(times[onsets], 0, 1, color='r', alpha=0.9, linestyle='--', label='Onsets')
    plt.title('Waveform and Detected Onsets')
    plt.xlabel('Time (s)')
    plt.legend()
    plt.savefig(f'{file_direction}_onset_detection.png')



import spacy
from epitran import Epitran
from pyphen import Pyphen
import re
import jiwer
import difflib
import logging
import time
from faster_whisper import WhisperModel
import multiprocessing
import concurrent.futures
import os

from .utils import (
    transcribe_file,
    split_audio_into_chunks
)


logging.basicConfig(format='APOLLO: (%(asctime)s): %(message)s', datefmt='%d/%m/%Y %I:%M:%S %p',
                    level=logging.DEBUG)


def clean_text(text: str) -> str:
    """
    Cleans the text from all punctuation, special characters, colons and semicolons
    it only leaves tildes and double points on top of words without changing the case

    :param text:
    :return:
    """
    try:
        cleaned_text = re.sub(r'[^A-Za-z0-9áéíóúÁÉÍÓÚñÑüÜ]+', ' ', text)
        return cleaned_text
    except Exception as e:
        raise Exception("Error getting clean text: {0}".format(e))


def sentences(text: str) -> list[str]:
    """
    Extracts all the sentences in a text string conserving inner punctuation and case

    :param text:
    :return:
    """
    try:
        sentence_delimiters = r'[.!?]|(?:\.{3})|…|¡|¿|;|"'
        spanish_prefixes = ['Sr.', 'Sra.', 'Dr.', 'Lic.', 'Ing.']

        for prefix in spanish_prefixes:
            text = text.replace(prefix, prefix.replace('.', '###'))

        sentences_ = re.split(sentence_delimiters, text)

        for prefix in spanish_prefixes:
            sentences_ = [sentence.replace(prefix.replace('.', '###'), prefix) for sentence in sentences_]

        processed_sentences = []
        for sentence in sentences_:
            sentence = sentence.strip()
            if sentence and not sentence.startswith("–") and not sentence.startswith(
                    "-"):
                sentence = sentence.strip(":")
                sentence = sentence.strip(", ")
                processed_sentences.append(sentence)

        return processed_sentences

    except Exception as e:
        raise Exception('Error getting sentences: {0}'.format(e))


def word_checker(original_text: str, new_text: str) -> dict:
    """
    Compares two transcriptions and returns the correct words, incorrect
    words and omitted words

    word error rate (WER)
    match error rate (MER)
    word information lost (WIL)
    word information preserved (WIP)

    :param original_text:
    :param new_text:
    :return:
    """
    correct_words = []
    omitted_words = []

    try:
        original_words = words(original_text, True)
        new_words = words(new_text, True)

        output = jiwer.process_words(clean_text(original_text), clean_text(new_text))

        wer = output.wer
        mer = output.mer
        wip = output.wip
        wil = output.wil

        for original_word in original_words:
            if original_word in new_words:
                new_words.remove(original_word)
                correct_words.append(original_word)
            else:
                omitted_words.append(original_word)

        incorrect_words = new_words

        return {"correct_words": correct_words, "omitted_words": omitted_words, "incorrect_words": incorrect_words,
                "mer": mer, "wil": wil, "wip": wip, "wer": wer}

    except Exception as e:
        raise Exception("Error comparing words: {0}".format(e))


def word_alignment(original_text: str, new_text: str) -> dict:
    """
    Compares two transcriptions and returns word by word the difference
    marked with "*" the length of the word and the residual words in order

    Residuals are marked with "+"
    Extras are marked with "*"

    :param original_text:
    :param new_text:
    :return:
    """
    original_words = words(original_text)
    new_words = words(new_text)
    matcher = difflib.SequenceMatcher(None, original_words, new_words)
    alignment = []
    residual_words = []
    extra_words = []

    try:
        for opcode, a0, a1, b0, b1 in matcher.get_opcodes():
            if opcode == 'equal':
                # If the words are the same, add them to the list
                alignment.extend(original_words[a0:a1])

            elif opcode == 'delete':
                # If the words are different, add asterisks to the list
                alignment.extend('+' * len(word) for word in original_words[a0:a1])
                residual_words.extend(original_words[a0:a1])

            elif opcode == 'insert' or opcode == 'replace':
                # Original transcription does not include this words
                alignment.extend('*' * len(word) for word in new_words[b0:b1])
                extra_words.extend(new_words[b0:b1])

        return {"alignment": alignment, "residual_words": residual_words, "extra_words": extra_words}
    except Exception as e:
        raise Exception("Error on word alignment: {0}".format(e))


def phonetic_transcription(text: str, lang='spa-Latn') -> str:
    """
    Cleans text and makes the phonetic following IPA

    :param text:
    :param lang:
    :return:
    """
    epi = Epitran(lang)
    try:
        return epi.transliterate(clean_text(text))
    except Exception as e:
        raise Exception("Error getting the phonetic transcription: {0}".format(e))


def syllables(text: str, lang='es') -> list[str]:
    """
    Extracts all the syllables from a text string

    :param text:
    :param lang:
    :return:
    """
    nlp = spacy.load('es_core_news_lg' if lang == 'es' else 'en_core_web_lg')
    py_instance = Pyphen(lang=lang)
    try:
        text = clean_text(text)
        doc = nlp(text)
        list_words = [token.text for token in doc]
        list_syllables = []

        for word in list_words:
            syllables_ = py_instance.inserted(word).split('-')
            list_syllables.extend(syllables_)

        return list_syllables

    except Exception as e:
        raise Exception('Error getting syllables: {0}'.format(e))


def words(text: str, lower=False, lang='es') -> list[str]:
    """
    Extracts all the words from a corpus in lower case

    :param text:
    :param lower:
    :param lang:
    :return:
    """
    nlp = spacy.load('es_core_news_lg' if lang == 'es' else 'en_core_web_lg')
    try:
        text = clean_text(text)

        if lower:
            doc = nlp(text.lower())
        else:
            doc = nlp(text)

        list_words = [token.text for token in doc]
        return list_words
    except Exception as e:
        raise Exception("Error on getting words: {0}".format(e))


def transcriber_parallel(file: str, model_name='small', device='cpu', max_processes=0, compute_type='float16',
                lang='es', silence_threshold: str = "-20dB", silence_duration: float = 2.0) -> dict:
    """
    Generates a transcription result from the given file
    it simplifies the audio loading and comes out of the box
    with all the timestamps of each word

    :param file: The name of the file
    :param model_name:
    :param device:
    :param max_processes:
    :param compute_type:
    :param lang:
    :param silence_threshold:
    :param silence_duration:
    :return:
    """
    if max_processes > multiprocessing.cpu_count() or max_processes == 0:
        max_processes = multiprocessing.cpu_count()

    temp_files_array = split_audio_into_chunks(file, max_processes, silence_threshold, silence_duration)
    model = WhisperModel(model_name, device=device, compute_type=compute_type, cpu_threads=4)
    segments_output = []
    futures = []

    try:
        logging.info("Starting transcription")
        start_time = time.time()

        with concurrent.futures.ThreadPoolExecutor(max_processes) as executor:
            for file_path in temp_files_array:
                future = executor.submit(transcribe_file, file_path, model, lang)
                futures.append(future)

        transcription = ""
        for future in futures:
            segments = future.result()

            for segment in segments:
                segments_output.append(
                    {"start": segment.start, "end": segment.end, "text": segment.text, "words": segment.words})

                transcription += segment.text + " "

        for temp_file in temp_files_array:
            os.remove(temp_file)

        return {"transcription": transcription, "segments": segments_output, "time_taken": time.time() - start_time}
    except Exception as e:
        raise Exception("Error transcribing: {0}".format(e))


def transcriber(file: str, model_name='large-v2', device='cuda', compute_type='float16', beam_size=5,
                lang='es') -> dict:
    """
    Generates a transcription result from the given file
    it simplifies the audio loading and comes out of the box
    with all the timestamps of each word

    :param file: The name of the file
    :param model_name:
    :param device:
    :param compute_type:
    :param beam_size:
    :param lang:
    :return:
    """

    model = WhisperModel(model_name, device=device, compute_type=compute_type)
    segments_output = []

    try:
        logging.info("Starting transcription")
        start_time = time.time()

        segments, info = model.transcribe(file, beam_size=beam_size, language=lang, word_timestamps=True)

        transcription = ""

        for segment in segments:
            segments_output.append(
                {"start": segment.start, "end": segment.end, "text": segment.text, "words": segment.words,
                 "tokens": segment.tokens})

            transcription += segment.text + " "

        return {"transcription": transcription, "segments": segments_output, "info": info, "time_taken": time.time() - start_time}
    except Exception as e:
        raise Exception("Error transcribing: {0}".format(e))



from __future__ import annotations
import sounddevice as sd
import numpy as np
import json
import keyboard
import librosa
import librosa.feature
import librosa.display
import matplotlib.pyplot as plt
import time
import logging
from moviepy.editor import VideoFileClip
from .utils import (
    lim_y_trim,
    ensure_dir_created,
    save_to_file,
    plot_chroma,
    plot_onsets,
    plot_spectral_contrast,
)

logging.basicConfig(format='APOLLO: (%(asctime)s): %(message)s', datefmt='%d/%m/%Y %I:%M:%S %p',
                    level=logging.DEBUG)


def record_voice(channels: int = 1, samplerate: int = 22050, save_audio=True, output_dir='./output',
                 filename='recording',
                 timestamp: float = None) -> dict:
    """
    Records the voice in a controlled timestamp, when pressing 's'
    or stop the execution it stops the recording

    It returns a dict with the time taken and the voice data as a Queue

    Also, it calls save_audio by default to ./output

    :param filename:
    :param channels:
    :param samplerate:
    :param save_audio:
    :param output_dir:
    :param timestamp:
    :return:
    """
    if timestamp and (timestamp > 30 or timestamp < 10):
        raise ValueError("Timestamp can not be longer than 30 or smaller than 10.")

    ensure_dir_created(output_dir)

    recording = True
    buffer_length = int(samplerate * timestamp) if timestamp else None
    audio_data = np.zeros((buffer_length, channels), dtype=np.float32) if buffer_length else []

    def callback(indata, _, __, ___) -> None:
        """
        A callback function to add data to the Queue current recording

        :param indata:
        :param _:
        :param __:
        :param ___:
        :return:
        """
        if recording:
            if buffer_length:
                audio_data[:indata.shape[0]] = indata
            else:
                audio_data.append(indata.copy())

    try:
        with sd.InputStream(callback=callback, channels=channels, samplerate=samplerate):
            logging.info("Recording... Press 's' or exit to stop.")
            start_time = time.time()
            while recording:
                if timestamp and time.time() - start_time >= timestamp:
                    recording = False
                    logging.info(f"Recording Stopped after {time.time() - start_time} seconds.")
                elif keyboard.is_pressed('s'):
                    recording = False
                    logging.info("Recording Stopped")

            if save_audio:
                save_to_file("{0}/{1}.wav".format(output_dir, filename), audio_data, samplerate)

        return {"time": time.time() - start_time, "voice": audio_data, "samplerate": samplerate}
    except Exception as e:
        raise Exception("Error recording voice: {0}".format(e))


def record_ontime(channels: int, samplerate: int, save_audio=True, output_dir='./output', filename='recording',
                  timestamp: int = 30, num_recordings: int = 5) -> list:
    """
    Record audios on time to generate data every timestamp time

    :param channels:
    :param samplerate:
    :param save_audio:
    :param output_dir:
    :param filename:
    :param timestamp:
    :param num_recordings:
    :return:
    """

    audios_data = []

    try:
        for i in range(num_recordings):
            result = record_voice(channels, samplerate, save_audio, output_dir, f"{filename}_{i}", timestamp)
            audios_data.append(result)

        return audios_data
    except Exception as e:
        raise Exception("Error recording audios: {0}".format(e))


def convert_to_audio(video_path: str, audio_path: str):
    """
    Converts any video file (mp4 preference) to audio file (mp3 preference)
    :param video_path:
    :param audio_path:
    :return:
    """
    audio_clip = None  # Initialize the variable
    try:
        video_clip = VideoFileClip(video_path)
        audio_clip = video_clip.audio
        audio_clip.write_audiofile(audio_path)
    except Exception as e:
        logging.error(f"An error occurred while extracting audio: {str(e)}")
    finally:
        if audio_clip is not None:
            audio_clip.close()  # Should not give an error
        if 'video_clip' in locals():
            video_clip.close()


def voice_intensity(filepath: str, lim: list[float, float] = None, plot=False, amplitude_env=False, frame_size=1024,
                    hop_length=512, output_dir='./output') -> str:
    """
    Generate audio data processed with librosa to create visualizations or audio processing
    if the lim is specified then the audio is cropped to that timestamp

    :param filepath: path to the audio file
    :param lim: list of two floats specifying start and end time for cropping (optional)
    :param plot: boolean flag to generate a plot (optional)
    :param amplitude_env: boolean flag to calculate the amplitude envelope (optional)
    :param frame_size: size of the STFT window (default: 1024)
    :param hop_length: hop length between consecutive STFT windows (default: 512)
    :param output_dir: directory to save the plot (default: ./output)
    :return: dictionary containing processed audio data
    """
    ensure_dir_created(output_dir)

    file_direction = output_dir + "/" + filepath.split("/")[-1].split(".")[0]

    try:
        y, sr = librosa.load(filepath)
        logging.info("Audio found, starting intensity processing...")
        duration = librosa.get_duration(y=y, sr=sr)

        if lim:
            y = librosa.util.sync(y, lim, sr=sr)

        # Calculate STFT and amplitude envelope
        stft = np.abs(librosa.stft(y, n_fft=frame_size, hop_length=hop_length))
        amplitude_envelope = np.array([max(y[i:i+frame_size]) for i in range(0, y.size, hop_length)])

        # Convert amplitude envelope to dB
        amplitude_envelope_db = librosa.amplitude_to_db(amplitude_envelope)

        max_amplitude = np.max(np.abs(y))

        data = {
            "duration": duration,
            "samplerate": sr,
            "max_amplitude": max_amplitude.tolist(),
            "stft": stft.tolist(),
            "amplitude_envelope_db": amplitude_envelope_db.tolist(),
        }

        if plot:
            try:
                frames = range(0, len(amplitude_envelope))
                t = librosa.frames_to_time(frames, sr=sr, hop_length=hop_length)
                plt.figure(figsize=(25, 10))
                librosa.display.waveshow(y, sr=sr)
                if amplitude_env:
                    plt.plot(t, amplitude_envelope, color="r", label="Amp_env")
                plt.xlabel('Time (seconds)')
                plt.ylabel('Audio Time Series')
                plt.title('Full Signal')
                plt.legend()
                plt.savefig(f'{file_direction}_full_signal.png')
            except Exception as e:
                raise Exception("Error saving plot: {0}".format(e))

    except Exception as e:
        raise Exception(f"An error occurred while processing audio: {str(e)}")

    return json.dumps(data)


def voice_expression(filepath: str, lim: list[float, float] = None, plot=False, output_dir='./output') -> str:
    """
    Generates voice expression data

    :param filepath:
    :param lim:
    :param plot:
    :param output_dir:
    :return:
    """
    ensure_dir_created(output_dir)

    file_direction = output_dir + "/" + filepath.split("/")[-1].split(".")[0]

    try:
        y, sr = librosa.load(filepath)
        logging.info("Audio found, starting expression processing...")

        if lim:
            y = lim_y_trim(sr, y, lim)

        # Chroma features
        n_chroma = 12
        n_fft = 2048
        hop_length = 512
        s = np.abs(librosa.stft(y, n_fft=n_fft)) ** 2
        chroma = librosa.feature.chroma_stft(S=s, sr=sr, n_chroma=n_chroma, n_fft=n_fft, hop_length=hop_length)

        # Spectral features
        n_bands = 6
        stft = np.abs(librosa.stft(y, n_fft=n_fft, hop_length=hop_length))
        contrast = librosa.feature.spectral_contrast(S=stft, sr=sr, n_bands=n_bands)

        # Onset detection
        onset_env = librosa.onset.onset_strength(S=stft, sr=sr)
        onsets = librosa.onset.onset_detect(onset_envelope=onset_env, sr=sr)
        times = librosa.times_like(onset_env, sr=sr)

        # Generate plots if requested
        if plot:
            plot_chroma(file_direction, s, chroma)
            plot_spectral_contrast(file_direction, contrast)
            plot_onsets(file_direction, onset_env, onsets, times)

        data = {
            "chroma": chroma.tolist(),
            "spectral_contrast": contrast.tolist(),
            "onsets": onsets.tolist(),
            "onset_envelope": onset_env.tolist(),
            "onset_times": times[onsets].tolist(),
        }

        return json.dumps(data)

    except Exception as e:
        raise Exception(f"Error processing voice: {e}")



import os
import ffmpeg
import tempfile
import re


def find_optimal_breakpoints(points: list[float], n: int) -> list[float]:
    """
    Generates a list of optimal breakpoints in a list of points.

    :param points: A list of points.
    :param n: The number of desired breakpoints.
    :return: A list of breakpoints.
    :raises ValueError: If `points` is empty or `n` is less than 1.
    """

    if not points:
        raise ValueError("points must not be empty")
    if n < 1:
        raise ValueError("n must be at least 1")

    result = []
    optimal_length = points[-1] / n
    current_point = 0
    last_breakpoint = 0

    for i in points[:-1]:
        if (i - last_breakpoint) >= optimal_length:
            if optimal_length - (current_point - last_breakpoint) < (i - last_breakpoint) - optimal_length:
                result.append(current_point)
            else:
                result.append(i)
            last_breakpoint = result[-1]
        current_point = i

    return result


def save_chunk_to_temp_file(input_file: str, start: float, end: float, suffix: str) -> str:
    temp_file = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)
    temp_file.close()

    in_stream = ffmpeg.input(input_file)
    (
        ffmpeg.output(in_stream, temp_file.name, ss=start, t=end - start, c="copy")
        .overwrite_output()
        .run()
    )

    return temp_file.name


def get_silence_starts(input_file: str, silence_threshold: str, silence_duration: str) -> list[float]:
    silence_starts = [0.0]

    reader = (
        ffmpeg.input(input_file)
        .filter("silencedetect", n=silence_threshold, d=silence_duration)
        .output("pipe:", format="null")
        .run_async(pipe_stderr=True)
    )

    silence_end_re = re.compile(
        r" silence_end: (?P<end>[0-9]+(\.?[0-9]*)) \| silence_duration: (?P<dur>[0-9]+(\.?[0-9]*))"
    )

    while True:
        line = reader.stderr.readline().decode("utf-8")
        if not line:
            break

        match = silence_end_re.search(line)
        if match:
            silence_end = float(match.group("end"))
            silence_dur = float(match.group("dur"))
            silence_start = silence_end - silence_dur
            silence_starts.append(silence_start)

    return silence_starts


def split_audio_into_chunks(input_file: str, max_chunks: int,
                            silence_threshold: str = "-20dB", silence_duration: float = 2.0) -> list[str]:
    """
    Splits an audio file into chunks based on periods of silence.

    :param input_file: The path of the input audio file.
    :param max_chunks: The maximum number of chunks to split the audio into.
    :param silence_threshold: The threshold for detecting silence, in dB.
    :param silence_duration: The minimum duration of silence to detect, in seconds.
    :return: A list of paths of the temporary files containing the audio chunks.
    :raises ValueError: If the input file doesn't exist or isn't a valid audio file, or if max_chunks is less than 1.
    """

    if not os.path.exists(input_file):
        raise ValueError("input_file must exist")
    if not os.path.isfile(input_file):
        raise ValueError("input_file must be a file")
    if max_chunks < 1:
        raise ValueError("max_chunks must be at least 1")

    file_extension = os.path.splitext(input_file)[1]
    metadata = ffmpeg.probe(input_file)
    duration = float(metadata["format"]["duration"])

    silence_starts = get_silence_starts(input_file, silence_threshold, str(silence_duration))
    silence_starts.append(duration)

    temp_files = []
    current_chunk_start = 0.0

    n = max_chunks
    selected_items = find_optimal_breakpoints(silence_starts, n)
    selected_items.append(duration)

    for j in range(0, len(selected_items)):
        temp_file_path = save_chunk_to_temp_file(input_file, current_chunk_start, selected_items[j], suffix=file_extension)
        temp_files.append(temp_file_path)

        current_chunk_start = selected_items[j]

    return temp_files


def transcribe_file(file_path, model, lang):
    segments, info = model.transcribe(file_path, language=lang, word_timestamps=True)
    segments = list(segments)
    return segments



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="apollo-ai",
    version="0.0.47",
    description="Framework to process 3 channels in one: Video, Audio & Text",
    package_dir={"": "apollo"},
    packages=find_packages(where="apollo"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/VerbaNexAI/APOLLO.AI",
    author="VerbaNex, Riddle",
    license="GPL-3.0",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "ctranslate2==3.20.0",
        "epitran==1.24",
        "faster-whisper==0.9.0",
        "ffmpeg-python==0.2.0",
        "huggingface-hub==0.17.3",
        "Jinja2==3.1.2",
        "jiwer==3.0.3",
        "keyboard==0.13.5",
        "librosa==0.10.1",
        "llvmlite==0.41.1",
        "matplotlib==3.7",
        "mediapipe==0.10.7",
        "nltk==3.8.1",
        "numpy==1.26.1",
        "opencv-contrib-python==4.8.1.78",
        "opencv-python==4.8.1.78",
        "pandas==2.1.2",
        "Pillow==10.1.0",
        "pyphen==0.14.0",
        "scikit-learn==1.3.2",
        "scipy==1.11.3",
        "sounddevice==0.4.6",
        "soundfile==0.12.1",
        "spacy==3.7.2",
        "spacy-legacy==3.0.12",
        "spacy-loggers==1.0.5",
        "sympy==1.12",
        "tqdm==4.66.1",
        "transformers==4.34.1",
        "sixdrepnet~=0.1.6",
        "setuptools~=65.5.1",
        "moviepy~=1.0.3"
    ],
    extras_require={
        "dev": ["pytest>=7.4.3", "twine>=4.0.2"]
    },
    python_requires=">=3.10"
)



import cv2
import pandas as pd
import matplotlib.pyplot as plt
from scipy import signal
from math import cos, sin
import numpy as np
import os


def filter_video_data(distances_data: list[list[float]], normalize: bool) -> pd.DataFrame:  # Name's not accurate
    """
    Not really sure yet what it does

    :param distances_data:
    :param normalize:
    :return:
    """
    try:
        df = pd.DataFrame(distances_data, columns=["Distance", "Timestamp"])
        df = df.sort_values(by=["Timestamp"])

        # Calculate the normalization based on min and max values
        if normalize:
            max_value = df['Distance'].max()
            min_value = df['Distance'].min()
            df['Distance'] = (df['Distance'] - min_value) / (max_value - min_value)

        df['Distance'] = signal.savgol_filter(df["Distance"], min(len(df["Distance"]), 51), 2)

        return df
    except Exception as e:
        raise Exception("Error normalizing the data: {0}".format(e))


def plot_pose_cube(img, yaw, pitch, roll, tdx=None, tdy=None, size=150.):
    """
    Plot a cube to the face, input is a cv2 image
    Where (tdx, tdy) is the translation of the face.
    For pose we have [pitch yaw roll tdx tdy tdz scale_factor]

    :param img:
    :param yaw:
    :param pitch:
    :param roll:
    :param tdx:
    :param tdy:
    :param size:
    :return:
    """

    p = pitch * np.pi / 180
    y = -(yaw * np.pi / 180)
    r = roll * np.pi / 180
    if tdx != None and tdy != None:
        face_x = tdx - 0.50 * size
        face_y = tdy - 0.50 * size

    else:
        height, width = img.shape[:2]
        face_x = width / 2 - 0.5 * size
        face_y = height / 2 - 0.5 * size

    x1 = size * (cos(y) * cos(r)) + face_x
    y1 = size * (cos(p) * sin(r) + cos(r) * sin(p) * sin(y)) + face_y
    x2 = size * (-cos(y) * sin(r)) + face_x
    y2 = size * (cos(p) * cos(r) - sin(p) * sin(y) * sin(r)) + face_y
    x3 = size * (sin(y)) + face_x
    y3 = size * (-cos(y) * sin(p)) + face_y

    # Draw base in red
    cv2.line(img, (int(face_x), int(face_y)), (int(x1), int(y1)), (0, 0, 255), 3)
    cv2.line(img, (int(face_x), int(face_y)), (int(x2), int(y2)), (0, 0, 255), 3)
    cv2.line(img, (int(x2), int(y2)), (int(x2 + x1 - face_x), int(y2 + y1 - face_y)), (0, 0, 255), 3)
    cv2.line(img, (int(x1), int(y1)), (int(x1 + x2 - face_x), int(y1 + y2 - face_y)), (0, 0, 255), 3)
    # Draw pillars in blue
    cv2.line(img, (int(face_x), int(face_y)), (int(x3), int(y3)), (255, 0, 0), 2)
    cv2.line(img, (int(x1), int(y1)), (int(x1 + x3 - face_x), int(y1 + y3 - face_y)), (255, 0, 0), 2)
    cv2.line(img, (int(x2), int(y2)), (int(x2 + x3 - face_x), int(y2 + y3 - face_y)), (255, 0, 0), 2)
    cv2.line(img, (int(x2 + x1 - face_x), int(y2 + y1 - face_y)),
             (int(x3 + x1 + x2 - 2 * face_x), int(y3 + y2 + y1 - 2 * face_y)), (255, 0, 0), 2)
    # Draw top in green
    cv2.line(img, (int(x3 + x1 - face_x), int(y3 + y1 - face_y)),
             (int(x3 + x1 + x2 - 2 * face_x), int(y3 + y2 + y1 - 2 * face_y)), (0, 255, 0), 2)
    cv2.line(img, (int(x2 + x3 - face_x), int(y2 + y3 - face_y)),
             (int(x3 + x1 + x2 - 2 * face_x), int(y3 + y2 + y1 - 2 * face_y)), (0, 255, 0), 2)
    cv2.line(img, (int(x3), int(y3)), (int(x3 + x1 - face_x), int(y3 + y1 - face_y)), (0, 255, 0), 2)
    cv2.line(img, (int(x3), int(y3)), (int(x3 + x2 - face_x), int(y3 + y2 - face_y)), (0, 255, 0), 2)

    return img


def plot_head_data_mov(file_direction: str, time_in_seconds: list[float], pitch_vals: list, yaw_vals: list,
                       roll_vals: list):
    """
    Plot the head movement over time in a 3d plot

    :param file_direction:
    :param time_in_seconds:
    :param pitch_vals:
    :param yaw_vals:
    :param roll_vals:
    :return:
    """
    plt.figure(figsize=(10, 8))

    off_screen_label_added = False

    plt.subplot(3, 1, 1)
    plt.plot(time_in_seconds, pitch_vals, label='Pitch')
    for i in range(len(pitch_vals)):
        if pitch_vals[i] == 0:
            plt.axvspan(time_in_seconds[i - 1], time_in_seconds[i], facecolor='red', alpha=0.3,
                        label='Off Screen' if not off_screen_label_added else "")
            off_screen_label_added = True
    plt.xlabel('Time (seconds)')
    plt.ylabel('Angle (degrees)')
    plt.legend()

    off_screen_label_added = False

    plt.subplot(3, 1, 2)
    plt.plot(time_in_seconds, yaw_vals, label='Yaw')
    for i in range(len(yaw_vals)):
        if yaw_vals[i] == 0:
            plt.axvspan(time_in_seconds[i - 1], time_in_seconds[i], facecolor='red', alpha=0.3,
                        label='Off Screen' if not off_screen_label_added else "")
            off_screen_label_added = True
    plt.xlabel('Time (seconds)')
    plt.ylabel('Angle (degrees)')
    plt.legend()

    off_screen_label_added = False

    plt.subplot(3, 1, 3)
    plt.plot(time_in_seconds, roll_vals, label='Roll')
    for i in range(len(roll_vals)):
        if roll_vals[i] == 0:
            plt.axvspan(time_in_seconds[i - 1], time_in_seconds[i], facecolor='red', alpha=0.3,
                        label='Off Screen' if not off_screen_label_added else "")
            off_screen_label_added = True
    plt.xlabel('Time (seconds)')
    plt.ylabel('Angle (degrees)')
    plt.legend()

    plt.tight_layout()
    plt.savefig(f'{file_direction}_head_movement.png')


def plot_head_data_speed(file_direction: str, time_in_seconds: np.ndarray, head_speeds_mm: np.ndarray,
                         threshold_mm: float):
    """
    Plot the head speed over time in mm/second

    :param file_direction:
    :param time_in_seconds:
    :param head_speeds_mm:
    :param threshold_mm:
    :return:
    """
    plt.figure()

    plt.plot(time_in_seconds, head_speeds_mm)
    off_screen_label_added = False

    # Checks for subject if it is not on-screen
    for i in range(len(head_speeds_mm)):
        if head_speeds_mm[i] == 0:
            plt.axvspan(time_in_seconds[i - 1], time_in_seconds[i], facecolor='red', alpha=0.3,
                        label='Off Screen' if not off_screen_label_added else "")
            off_screen_label_added = True

    plt.xlabel('Time (seconds)')
    plt.ylabel('Head Movement Speed (mm)')
    plt.title('Head Movement Speed Over Time')
    plt.axhline(y=threshold_mm, color='r', linestyle='--', label='Threshold')
    plt.legend()
    plt.savefig(f'{file_direction}_head_speed.png')


def ensure_dir_created(output_dir: str) -> None:
    """
    Util function to ensure directory has been created and if
    not then create such directory

    :param output_dir:
    :return:
    """
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
